#![allow(clippy::not_unsafe_ptr_arg_deref)]

use libc::{
    c_int, c_uchar, c_uint, c_void, dlsym, getrandom as libc_getrandom, ssize_t, RTLD_NEXT,
};
use std::env;
use std::ffi::{CStr, CString};
use std::mem;
use std::sync::OnceLock;

fn fakerand_value() -> Option<i32> {
    static FAKERAND: OnceLock<Option<i32>> = OnceLock::new();
    *FAKERAND.get_or_init(|| env::var("FAKERAND").ok().and_then(|v| v.parse().ok()))
}

#[no_mangle]
pub extern "C" fn rand() -> c_int {

    if let Some(fakerand) = fakerand_value() {
        println!("rand() value {}", fakerand);
        return fakerand;
    }

    // match fakerand_value() {
    //     Some(n) => return n,
    //     None => {
    //         eprint!("fakerand loaded but unable to generate fake rand");
    //         unsafe {
    //             let errno = libc::__errno_location();
    //             *errno = libc::EINVAL;
    //         }
    //         return -1;
    //     }
    // };

    unsafe {
        let symbol = dlsym(RTLD_NEXT, "rand\0".as_ptr() as *const _);

        if !symbol.is_null() {
            let original_rand: extern "C" fn() -> c_int = std::mem::transmute(symbol);
            original_rand()
        } else {
            eprintln!("Error: original rand() not found");
            std::process::exit(1);
        }
    }
}

#[inline(always)]
unsafe fn copy_slice_to_ptr(slice: &[u8], buf: *mut c_void, buf_size: usize) {
    for offset in (0..buf_size).step_by(4) {
        unsafe {
            std::ptr::copy(
                slice.as_ptr() as *mut c_void,
                buf.byte_add(offset),
                std::cmp::min(buf_size - offset, slice.len()),
            )
        }
    }
}

/// # Safety
///
/// This function assumes the `read()` libc function is defined as
/// `read(c_int, *mut c_void, usize) -> ssize_t` and calls it as such. This function also invokes
/// `libc::readlink()` with two paths each of which not exceeding 4096 (PATH_MAX) size in length.
///
/// The function may panic if the path can't be encoded as a Rust `str` value.
#[no_mangle]
pub extern "C" fn read(fd: c_int, buf: *mut c_void, count: usize) -> ssize_t {
    static ORIGINAL_READ: OnceLock<extern "C" fn(c_int, *mut c_void, usize) -> ssize_t> =
        OnceLock::new();

    let original_read = ORIGINAL_READ.get_or_init(|| {
        let symbol = unsafe { dlsym(RTLD_NEXT, "read\0".as_ptr() as *const _) };
        if symbol.is_null() {
            eprintln!("Error: original read() not found");
            std::process::exit(1);
        }
        unsafe { std::mem::transmute(symbol) }
    });

    let path = format!("/proc/self/fd/{}", fd);
    // PATH_MAX is defined as 4096 chars in a path including nul
    let mut actualpath = [0u8; 4096];

    let len = unsafe {
        libc::readlink(
            path.as_ptr() as *const i8,
            actualpath.as_mut_ptr() as *mut i8,
            actualpath.len() as _,
        )
    };

    if len != -1 {
        let actualpath = unsafe { CStr::from_ptr(actualpath.as_ptr() as *const i8) };
        let urandom = CString::new("/dev/urandom").unwrap();
        let random = CString::new("/dev/random").unwrap();
        if actualpath == urandom.as_c_str() || actualpath == random.as_c_str() {
            let random_value = match fakerand_value() {
                Some(n) => n,
                None => {
                    eprintln!("fakerand loaded but unable to generate fake rand");
                    unsafe {
                        let errno = libc::__errno_location();
                        *errno = libc::EINVAL;
                    }
                    return -1;
                }
            };
            unsafe {
                copy_slice_to_ptr(&random_value.to_ne_bytes(), buf, count);
            }
            /*
            let value_ptr: *const i32 = &random_value;
            for offset in (0..count).step_by(4) {
                unsafe {
                    std::ptr::copy(
                        value_ptr as *mut c_void,
                        buf.byte_add(offset),
                        std::cmp::min(count - offset, 4),
                    )
                };
            }
            */
            return count.try_into().expect("size larger than SSIZE_MAX");
        }
        /*
        if (actualpath == "/dev/urandom" || actualpath == "/dev/random")
            && fakerand_value().is_some()
        {
            let fakerand = fakerand_value().unwrap();
            *(buf as *mut c_int) = fakerand;
            return std::mem::size_of::<c_int>() as ssize_t;
        }
        */
    }

    original_read(fd, buf, count)
}

/// # Safety
///
/// Checked by caller. This function may result in undefined behavior if buflen is longer than the
/// allocated length of buf or if buf is not allocated, or previously freed, memory.
#[no_mangle]
pub extern "C" fn getrandom(buf: *mut c_void, buflen: usize, flags: c_uint) -> ssize_t {
    if let Some(fakerand) = fakerand_value() {
        unsafe {
            *(buf as *mut c_int) = fakerand;
        }
        return std::mem::size_of::<c_int>() as ssize_t;
    }

    unsafe { libc_getrandom(buf, buflen, flags) }
}

/// # Safety
///
/// Checked by caller. This function may result in undefined behavior if num is longer than the
/// allocated length of buf or if buf is not allocated, or previously freed, memory.
#[no_mangle]
pub extern "C" fn RAND_bytes(buf: *mut c_uchar, num: c_int) -> c_int {
    if let Some(fakerand) = fakerand_value() {
        unsafe {
            for i in 0..num {
                *buf.add(i as usize) = fakerand as c_uchar;
            }
        }
        return 1;
    }

    unsafe {
        let symbol = dlsym(RTLD_NEXT, "RAND_bytes\0".as_ptr() as *const _);
        if !symbol.is_null() {
            let original_rand_bytes: extern "C" fn(*mut c_uchar, c_int) -> c_int =
                mem::transmute(symbol);
            original_rand_bytes(buf, num)
        } else {
            eprintln!("Error: original RAND_bytes() not found");
            std::process::exit(1);
        }
    }
}

#[cfg(test)]
mod tests {
    use std::fs::OpenOptions;
    use std::os::unix::io::AsRawFd;

    use super::*;

    #[test]
    fn test_rand_with_fakerand() {
        env::set_var("FAKERAND", "42");
        assert_eq!(rand(), 42);
        env::remove_var("FAKERAND");
    }

    #[test]
    fn test_rand_without_fakerand() {
        env::remove_var("FAKERAND");
        let rand_value = rand();
        // TODO: Does this mean the test will fail once in a while?
        assert!(rand_value != 42);
        assert!(rand_value >= 0 && rand_value <= libc::RAND_MAX);
    }

    #[test]
    fn test_read_with_fakerand() {
        env::set_var("FAKERAND", "42");

        let file = OpenOptions::new()
            .read(true)
            .open("/dev/urandom")
            .expect("Failed to open /dev/urandom");
        let fd = file.as_raw_fd();
        let mut buf = [0u8; 4];
        let result = read(fd, buf.as_mut_ptr() as *mut c_void, buf.len());

        assert_eq!(result, std::mem::size_of::<c_int>() as ssize_t);
        assert_eq!(buf, [42, 0, 0, 0]);

        env::remove_var("FAKERAND");
    }

    #[test]
    fn test_read_without_fakerand() {
        env::remove_var("FAKERAND");

        let file = OpenOptions::new()
            .read(true)
            .open("/dev/urandom")
            .expect("Failed to open /dev/urandom");
        let fd = file.as_raw_fd();
        let mut buf = [0u8; 4];
        let result = read(fd, buf.as_mut_ptr() as *mut c_void, buf.len());

        assert_eq!(result as usize, buf.len());
        assert!(buf != [0, 0, 0, 0]);
    }

    #[test]
    fn test_rand_bytes_with_fakerand() {
        env::set_var("FAKERAND", "42");

        let mut buf = [0u8; 4];
        let result = RAND_bytes(buf.as_mut_ptr(), buf.len() as c_int);

        assert_eq!(result, 1);
        assert_eq!(buf, [42, 42, 42, 42]);

        env::remove_var("FAKERAND");
    }
}
